﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _____________________________________________________                               //
//                                                                                      //
//  Copyright © 2016 Reimar Grasbon (softwork@grasbon.de)                               //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "Timer.h"
#include "String.h"
#include "Display.h"
#include "Sound.h"
#include "Board.h"
#include "Storing.h"
#include "Alarm.h"
#include "Menu.h"
#include "Application.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum       // Zustände des Radios:
{   RS_OFF,        // · Ist  ausgeschaltet
    RS_ENABLING,   // · Wird eingeschaltet
    RS_ON,         // · Ist  eingeschaltet
    RS_DISABLING   // · Wird ausgeschaltet
} RadioState;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static OperationMode mode            = OM_CLOCK;  // Betriebsmodus
static RadioState    radio           = RS_OFF;    // Zustände des Radios
static UInt08        wakeup          = true;      // Auslöser: Anzeige einschalten
static bool          keyPressed      = false;     // 'true' = Tasten gedrückt Ereignis
static SInt08        encoderPosition = 0;         // Drehinkrementalgeber-Position

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen   //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  GetLocalDateTime(): Die Systemzeit wird vom Board-Modul in UTC zur Verfügung ge-	//
//  ------------------  stellt. Diese Funktion liest die Systemzeit aus und versetzt	//
//	sie unter Berücksichtigung der Sommerzeiteinstellung in die lokale Zeitzone. Ist	//
//	die Systemuhr nicht gestellt, ist die  zurückgelieferte Datumzeit ungültig.			//
//																						//
//  Eingabe: · dateTime: Datumzeit, die die lokale Zeit und das Datum aufnimmt.			//
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
GetLocalDateTime(DateTime* dateTime)
{
    BOARD_Clock(dateTime);
    DaylightSavingMode daylightSavingMode;
    STORING_Load(STO_DaylightSavingMode, &daylightSavingMode);
    switch( daylightSavingMode )
    {   case DSM_OFF:  DATETIME_Localize(CFG_TIMEZONE  , false, dateTime); break;
        case DSM_ON:   DATETIME_Localize(CFG_TIMEZONE+1, false, dateTime); break;
        case DSM_AUTO: DATETIME_Localize(CFG_TIMEZONE  , true , dateTime);
    }    
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsHoliday(): Prüft, ob es sich bei dem angegebenen Tag um einen Feiertag handelt.   //
//  -----------  Die Regelung der Feiertage ist von Bundesland zu Bundesland unter-     //
//  schiedlich. Hier folgende Festlegung:                                               //
//                                                                                      //
//                                                                                      //
//         Feiertage           |BW|BY|BE|BB|HB|HH|HE|MV|NI|NW|RP|SL|SN|ST|SH|TH|        //
//       ----------------------+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+        //
//         Neujahr             |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Heilig-Dreikönige   |**|**|  |  |  |  |  |  |  |  |  |  |  |**|  |  |        //
//         Frauentag           |  |  |**|  |  |  |  |  |  |  |  |  |  |  |  |  |        //
//         Karfreitag          |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Ostersonntag        |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Ostermontag         |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Tag der Arbeit      |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Christi Himmelfahrt |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Pfingstsonntag      |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Pfingstmontag       |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Fronleichnam        |**|**|  |  |  |  |**|  |  |**|**|**|  |  |  |  |        //
//         Mariä Himmelfahrt   |  |**|  |  |  |  |  |  |  |  |  |**|  |  |  |  |        //
//         Weltkindertag       |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |**|        //
//         Tag der dt. Einheit |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         Reformationstag     |  |  |  |**|**|**|  |**|**|  |  |  |**|**|**|**|        //
//         Allerheiligen       |**|**|  |  |  |  |  |  |  |**|**|**|  |  |  |  |        //
//         Buß- und Bettag     |  |  |  |  |  |  |  |  |  |  |  |  |**|  |  |  |        //
//         1. Weihnachtstag    |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//         2. Weihnachtstag    |**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|**|        //
//       ----------------------+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+        //
//                                                                                      //
//                                                                                      //
//  Eingabe: · region:  Deutsches Bundesland als Aufzählungstyp 'GermanRegion'.         //
//           · dayType: Tagestyp gem. Aufzählungstyp 'DTDayType'.                       //
//  Ausgabe: 'true', wenn der angegebene Tag im gewählten Bundesland ein Feiertag ist.  //
//////////////////////////////////////////////////////////////////////////////////////////
bool
IsHoliday(GermanRegion region, DTDayType dayType)
{
    static const DTDayType  holidayAll[] PROGMEM =
    {   // Bundesländer übergreifende Feiertage
        DT_NEW_YEAR,
        DT_GOOD_FRIDAY,
        DT_EASTER_SUNDAY,
        DT_EASTER_MONDAY,
        DT_MAY_DAY,
        DT_ASCENSION_THURSDAY,
        DT_WHITSUNDAY,
        DT_WHITMONDAY,
        DT_OCTOBER_3RD,
        DT_CHRISTMAS_DAY,
        DT_BOXING_DAY
    };
    static const DTDayType  holidaySpec[][4] PROGMEM =
    {   // Bundesländer eigene Feiertage
        {   DT_EPIPHANY, DT_CORPUS_CHRISTI, DT_ALLHALLOWS, 0                 }, // · BW
        {   DT_EPIPHANY, DT_CORPUS_CHRISTI, DT_ASSUMPTION_DAY, DT_ALLHALLOWS }, // · BY
        {   DT_WOMEMS_DAY, 0, 0, 0                                           }, // · BE
        {   DT_REFORMATION_DAY, 0, 0, 0                                      }, // · BB
        {   DT_REFORMATION_DAY, 0, 0, 0                                      }, // · HB
        {   DT_REFORMATION_DAY, 0, 0, 0                                      }, // · HH
        {   DT_CORPUS_CHRISTI, 0, 0, 0                                       }, // · HE
        {   DT_REFORMATION_DAY, 0, 0, 0                                      }, // · MV
        {   DT_REFORMATION_DAY, 0, 0, 0                                      }, // · NI
        {   DT_CORPUS_CHRISTI, DT_ALLHALLOWS, 0, 0                           }, // · NW
        {   DT_CORPUS_CHRISTI, DT_ALLHALLOWS, 0, 0                           }, // · RP
        {   DT_CORPUS_CHRISTI, DT_ASSUMPTION_DAY, DT_ALLHALLOWS, 0           }, // · SL
        {   DT_REFORMATION_DAY, DT_REPENTANCE_DAY, 0, 0                      }, // · SN
        {   DT_EPIPHANY, DT_REFORMATION_DAY, 0, 0                            }, // · ST
        {   DT_REFORMATION_DAY, 0, 0, 0                                      }, // · SH
        {   DT_CHILDRENS_DAY, DT_REFORMATION_DAY, 0, 0                       }  // · TH
    };

    if( dayType != DT_NORMAL_WORKDAY && dayType != DT_NORMAL_WEEKEND )
    {   for( UInt08 i=0; i<CARDINALITY(holidayAll); i++ )
        {   if( dayType == pgm_read_byte(&holidayAll[i]) ) return true;
        }
        for( UInt08 i=0; i<CARDINALITY(holidaySpec[0]); i++ )
        {   if( dayType == pgm_read_byte(&holidaySpec[region][i]) ) return true;
        }
    }
    return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  IsOffDay(): Prüft, ob das angegebene Datum ein (arbeits)freier Tag ist. Das ist		//
//  ----------  der Fall, wenn 'date' ein Samstag, Sonntag oder ein Feiertag ist.		//
//																						//
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', der der Tag in 'date' ein freier Tag ist.							//
//////////////////////////////////////////////////////////////////////////////////////////
static bool
IsOffDay(const Date* date)
{
	DTWeekday weekday   = DATE_WeekDay(date);
    bool      isWeekEnd = weekday == DT_SAT || weekday == DT_SUN;	
    bool      isHoliday = APPLICATION_IsHoliday(DATE_DayType(date));
	return isWeekEnd || isHoliday;	
}

//////////////////////////////////////////////////////////////////////////////////////////
//  EnterReceiveMode(): Einstiegsfunktion der Betriebsart 'OM_RECEIVING'. Die Funktion 	//
//  ------------------  wird immer dann aufgerufen, wenn von einer anderen Betriebsart	//
//	in 'OM_RECEIVING' gewechselt wird.													//
//																						//
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterReceiveMode(void)
{
    DISPLAY_SetBlinkingDigits(0b01000010);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedReceiveMode(): Verfahrensfunktionen der Betriebsart 'OM_RECEIVING'. In       //
//  --------------------  diesem Modus wird der Empfangsfortschritt der Funkzeit an-	//
//	gezeigt. Sobald die komplette Zeitinformation empfangen wurde, wird zur Betriebs-	//
//	art 'OM_CLOCK' gewechselt.															//
//																						//
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedReceiveMode(void)
{
    // Empfangsfortschritt
    DateTime dateTime;
    GetLocalDateTime(&dateTime);
    if( DATETIME_IsValid(&dateTime) )
    {   // Gültige Zeit erhalten
        DISPLAY_StartAnim(ANIM_TWISTING_DIGITS);
        SOUND_Start(SOUND_TWISTING_DIGITS);
        bool alarmEnabled;
        STORING_Load(STO_AlarmEnabled, &alarmEnabled);
        if( alarmEnabled ) ALARM_Enable(true);
        APPLICATION_ChangeModeTo(OM_CLOCK);
    }
    else
    {   // Bitnummer (= Sekunde) anzeigen
        String string = STRING;
        STRING_PrintLit(&string, " --");
        UInt08 second = BOARD_DCFStatus()->bitCount;
        if( second < 10 ) STRING_PrintChar(&string, '0');
        STRING_PrintNum(&string, second);
        STRING_PrintLit(&string, "-- ");
        DISPLAY_Print(&string);
        DISPLAY_SetSegment(2, 0b0110001);
        DISPLAY_SetSegment(5, 0b0000111);
    }
    if( BOARD_KeyTime() > CFG_ENTER_SETTINGS_TIME ) APPLICATION_ChangeModeTo(OM_SETTINGS);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  EnterClockMode(): Einstiegsfunktion der Betriebsart 'OM_CLOCK'. Die Funktion wird	//
//  ----------------  immer dann aufgerufen, wenn von einer anderen Betriebsart	in      //
//	'OM_CLOCK' gewechselt wird.															//
//																						//
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterClockMode(void)
{
    DISPLAY_SetBlinkingDigits(0);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedClockMode(): Verfahrensfunktionen der Betriebsart 'OM_CLOCK'. 'OM_CLOCK' ist	//
//  ------------------  die Standard-Betriebsart, die zur Darstellung der lokalen		//
//	Uhrzeit genutzt wird. Nur in dieser Betriebsart ist es möglich, den Wecker zu		//
//	stellen ('OM_MENU')	oder in die Geräteeinstellungen zu wechseln ('OM_SETTINGS').	//
//	Ist die aktuelle Zeit nicht gültig, wird in die Betriebsart 'OM_RECEIVING' ge-		//
//	wechselt.																			//
//																						//
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedClockMode(void)
{
    // Zeitanzeige
    DateTime dateTime;
    GetLocalDateTime(&dateTime);
    if( DATETIME_IsValid(&dateTime) )
	{	String string = STRING;
        if( dateTime.time.hour < 10 ) STRING_PrintChar(&string, ' ');
        STRING_PrintNum(&string, dateTime.time.hour);
        STRING_PrintChar(&string, '-');
        if( dateTime.time.minute < 10 ) STRING_PrintChar(&string, '0');
        STRING_PrintNum(&string, dateTime.time.minute);
        if( ALARM_IsAlarm() )
        {   STRING_PrintChar(&string, ' ');
            if( ALARM_IsSnooze() )
            {   static const UInt08 anim[][2] PROGMEM =
                {   { 0b1000000, 0b1000000 },
                    { 0b0000000, 0b1100000 },
                    { 0b0000000, 0b0110000 },
                    { 0b0000000, 0b0011000 },
                    { 0b0001000, 0b0001000 },
                    { 0b0001100, 0b0000000 },
                    { 0b0000110, 0b0000000 },
                    { 0b1000010, 0b0000000 }
                };
                static Timer timer = TIMER(TIMER_MILLISECONDS, 150);
                if( TIMER_RestartIfExpired(&timer) )
                {   static UInt08 frame = 0;
                    DISPLAY_SetSegment(6, pgm_read_byte(&anim[frame][0]));
                    DISPLAY_SetSegment(7, pgm_read_byte(&anim[frame][1]));
                    frame++;
                    if( frame == CARDINALITY(anim) ) frame = 0;
                }
            }
        }
        else
        {   STRING_PrintChar(&string, '-');
            if( dateTime.time.second < 10 ) STRING_PrintChar(&string, '0');
            STRING_PrintNum(&string, dateTime.time.second);
        }
        DISPLAY_Print(&string);
    }
    else
    {   APPLICATION_ChangeModeTo(OM_RECEIVING);
    }
    if( !ALARM_IsAlarm() )
    {   if( BOARD_KeyTime() > CFG_ENTER_SETTINGS_TIME )
		{	APPLICATION_ChangeModeTo(OM_SETTINGS);
		}
        else if( APPLICATION_Encoder() != 0 )           
		{	APPLICATION_ChangeModeTo(OM_MENU);
		}
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedUserInteraction(): Verfahrensfunktion der Anwender-Eingaben. Die Funktion	//
//  ------------------------  fragt die Taste und den Drehinkrementalgeber der Konsole	//
//	ab und stellt deren Zustände in 'keyPressed' und 'encoderPosition' zur Verfügung.	//
//	Die Ereignisse werden bei jedem Aufruf neu ermittelt, d.h. die Werte stehen nur für //
//	einen Durchlauf der Hauptschleifen zur Verfügung.									//
//																						//
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedUserInteraction(void)
{
    // Nutzerinteraktion holen und bereitstellen
    keyPressed      = BOARD_TakeKey();
    encoderPosition = BOARD_TakeEncoder();
    if( keyPressed || encoderPosition != 0 ) APPLICATION_WakeDisplay();
    // Bei Abschluss einer Animationssequenz Nutzerinteraktion löschen
    static bool  animRunning = false;
    if( animRunning && DISPLAY_CurrentAnim() == ANIM_NONE )
    {   keyPressed = false;
        encoderPosition = 0;
    }
    animRunning = DISPLAY_CurrentAnim() != ANIM_NONE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedDisplaySaving(): Verfahrensfunktion der Anzeigen-Abschaltung. Der Automat    //
//  ----------------------  hat folgendes Zustandsdiagramm:                             //
//                                                                                      //
//                                            Zustände:                                 //
//                                             · S0: Anzeige ist abgeschaltet           //
//                   +--------+                · S1: Anzeige ist eingeschaltet          //
//    AC,DISABLE()   |////////| ON,ENABLE()    · S2: Anzeige wird abgeschaltet          //
//         +-------->|// S0 //|---------+                                               //
//         |         |////////|         |     Ereignisse (in absteigender Prorität):    //
//         |         +--------+         |      · ON: Einschalt-Auslöser                 //
//         |                            |      · TO: Zeitgeber abgelaufen               //
//    +----+---+                        |      · AC: Animation abgelaufen               //
//    |////////| ON,RECOVER()           |                                               //
//    |// S2 //+--------+               |     Aktionen:                                 //
//    |////////|        |   ON,TIMER()  |      · ENABLE():    Anzeige einschalten und   //
//    +--------+        |  +--------+   |                     Zeitgeber starten         //
//        /|\          \|/ |        |   |      · DISABLE():   Anzeige abschalten        //
//         |         +-----+--+     |   |      · TIMER():     Zeitgeber starten         //
//         |         |////////|<----+   |      · ANIMATION(): Animation starten und     //
//         +---------+// S1 //|         |                     Zeitgeber abschalten      //
//   TO,ANIMATION()  |////////|<--------+      · RECOVER():   Animation anhalten und    //
//                   +--------+                               Zeitgeber wieder starten  //
//                                             · NONE():      Keine Aktion              //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedDisplaySaving(void)
{
    typedef enum
    {   S0 = 0, S1 = 1, S2 = 2
    } State;

    enum   // Ereignisse
    {   ON = BV(0),
        TO = BV(1),
        AC = BV(2)
    };

    typedef enum
    {   ENABLE, DISABLE, TIMER, ANIMATION, RECOVER, NONE
    } Action;

    typedef struct
    {   State   state;
        Action  action;
    } Entry;

    static const Entry zeta[][3] PROGMEM =
    {   //+---- S0 ----+,+----- S1 ------+,+----- S2 -----+
        { { S1, ENABLE },{ S1, TIMER     },{ S1, RECOVER  } }, // ON
        { { S0, NONE   },{ S2, ANIMATION },{ S2, NONE     } }, // TO
        { { S0, NONE   },{ S1, NONE      },{ S0, DISABLE  } }, // AC
    };
    // Anzeige einschalten ausgelöst durch Bewegungsmelder
    {   static Timer update = TIMER(TIMER_MILLISECONDS, 50);
        if( TIMER_RestartIfExpired(&update) )
        {   static SInt08 sensitivity = 0;
            if( BOARD_PIRSensor() )
            {   SInt08 sensitivityLevel;
                STORING_Load(STO_MotionSensDisplay, &sensitivityLevel);
                sensitivityLevel = 9 - sensitivityLevel;
                if( sensitivity < sensitivityLevel*8 ) sensitivity++;
                else                                   wakeup = true;
            }
            else
            {   if( sensitivity > 0 ) sensitivity--;
            }
        }
    }
    // Ereigniserzeugung
    UInt08  eventSet = 0;
    // ON: Einschalt-Auslöser
    if( mode == OM_MENU || mode == OM_SETTINGS ) eventSet |= ON;    // Betriebsmodus
    if( ALARM_AnnouncementPhase() )              eventSet |= ON;    // Weckankündigung
    if( wakeup )                                 eventSet |= ON;    // Explizit
    wakeup = false;
    // TO: Zeitgeber abgelaufen
    static Timer timer = TIMER(TIMER_SECONDS, 0);
    if( TIMER_HasExpired(&timer) ) eventSet |= TO;
    // AC: Animation abgelaufen
    if( DISPLAY_CurrentAnim() != ANIM_GOTO_SLEEP ) eventSet |= AC;
    // Zustandsübergang
    static State state = S0;
    Action action = NONE;
    for( UInt08 i=0; i<CARDINALITY(zeta); i++ )
    {   const Entry* entry = &zeta[i][(UInt08)state];
        if( (eventSet & BV(i)) != 0 )
        {   State newState = (State)(pgm_read_byte(&entry->state));
            Action tmpAction = (Action)(pgm_read_byte(&entry->action));
            if( state == newState && tmpAction == NONE ) continue;
            state  = newState;
            action = tmpAction;
            break;
        }
    }
    // Aktionsausführung
    switch( action )
    {   case ENABLE:
        {   DISPLAY_Enable(true);
            UInt08 timeout;
            STORING_Load(STO_DisplayBlankStart, &timeout);
            TIMER_Start(&timer, timeout*60);
            break;
        }
        case DISABLE:
        {   DISPLAY_Enable(false);
            break;
        }
        case TIMER:
        {   UInt08 timeout;
            STORING_Load(STO_DisplayBlankStart, &timeout);
            TIMER_Start(&timer, timeout*60);
            break;
        }
        case ANIMATION:
        {   DISPLAY_StartAnim(ANIM_GOTO_SLEEP);
            break;
        }
        case RECOVER:
        {   DISPLAY_StopAnim();
            UInt08 timeout;
            STORING_Load(STO_DisplayBlankStart, &timeout);
            TIMER_Start(&timer, timeout*60);
            break;
        }
        default:
            ;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedDisplayBrightness(): Verfahrensfunktion der Anzeigen-Helligkeitssteuerung.	//
//  --------------------------															//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedDisplayBrightness(void)
{
    if( mode == OM_CLOCK )
    {   static Timer timer = TIMER(TIMER_MILLISECONDS, (10-CFG_DISPLAY_BRIGHT_AGILITY)*300);
        if( !TIMER_RestartIfExpired(&timer) ) return;
        SInt08 ambBrightness = BOARD_AmbientBrightness(); // Helligkeit in 0 .. 100%
        if( ambBrightness < 0 ) ambBrightness = 100;
        ambBrightness = (SInt08)(((SInt16)ambBrightness*100+333)/666); // % -> 0 .. 15
        SInt08 ledBrightness = DISPLAY_Brightness();
        if     ( ambBrightness < ledBrightness ) DISPLAY_SetBrightness(ledBrightness-1);
        else if( ambBrightness > ledBrightness ) DISPLAY_SetBrightness(ledBrightness+1);
    }
    else
    {   if( DISPLAY_Brightness() != 15 ) DISPLAY_SetBrightness(15);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedDotSignaling(): Verfahrensfunktion der Punkt-Signalisierung. Einige System-	//
//  ---------------------  zustände und auch Fehler werden durch exponierte LEDs in		//
//	der Anzeige dargestellt. Diese Funktion kümmert sich um die Signalisierungen.		//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedDotSignaling(void)
{
    // Wecker-Ein Anzeige
    {   bool on = false;
        if( ALARM_IsAlarm() ) on = BOARD_SecTicks()&1;
        else                  on = ALARM_IsSet();
        DISPLAY_SetDot(3, on);
    }
    // Fehlersignalisierung durch Blinkimpulse der Fehler-LED
    {   bool on = false;
        if( BOARD_RTClockFailure() )
        {   // RTC-Fehler. Signal: *-*-*-*- ...
            on = !!(BOARD_MSecTicks() & 0x0100);
        }
        else if( BOARD_DCFStatus()->elapsed >= CFG_DCF_FAIL_TIME )
        {   // DCF-Funkausfall. Signal: ----*-*----*-*----*-*--...
            UInt16 msecTicks = BOARD_MSecTicks();
            on = !!((msecTicks&0x0380)==0x0080 || (msecTicks&0x0380)==0x0180);
        }
        DISPLAY_SetDot(4, on);
    }
    // Markierung Wochenend/Feiertag im Dezimalpunkt der rechtesten Stelle
    {   bool on = false;
        bool showHoliday;
        STORING_Load(STO_ShowHoliday, &showHoliday);
        if( showHoliday )
        {   DateTime  dateTime;
            GetLocalDateTime(&dateTime);
			on = IsOffDay(&dateTime.date);            
        }
        DISPLAY_SetDot(7, on);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedRadioSSR(): Verfahrensfunktion der direkten Relaisschaltung (Radio) und der 	//
//  -----------------  Steuerung der zugehörigen LED auf der Konsole.					//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedRadioSSR(void)
{
    if( ALARM_IsAlarm() )
    {   // Weckalarm: LED auf der Konsole blinken lassen
		bool on = BOARD_SecTicks()&1;
        if( on ) BOARD_SetLedBrightness(100);
        else     BOARD_SetLedBrightness(  0);
        return;
    }
	if( APPLICATION_Key() )
	{	// Manuelles ein-/ausschalten des Radios
		if( mode == OM_CLOCK || mode == OM_RECEIVING )
		{   if( DISPLAY_CurrentAnim() == ANIM_NONE )        
            {   if( radio == RS_OFF ) radio = RS_ENABLING;
                if( radio == RS_ON  ) radio = RS_DISABLING;
            }
        }
    }                                               //             +--------+
    static SInt08 rising = true;                    //        +--->| RS_OFF |---+
    switch( radio )                                 //        |    +--------+   |
    {   case RS_ENABLING:                           //        |                \|/
            //Radio wird eingeschaltet              // +------+-------+  +-------------+
            SOUND_Start(SOUND_RADIO_ON);            // | RS_DISABLING |  | RS_ENABLING |
            DISPLAY_StartAnim(ANIM_RADIO_ON);       // +--------------+  +------+------+
            BOARD_SetLedBrightness(0);              //       /|\                |
            rising = true;                          //        |    +-------+    |
            radio  = RS_ON;                         //        +----| RS_ON |<---+
            break;                                  //             +-------+
        case RS_ON:
            // Radio ist  eingeschaltet
            BOARD_EnableRelais(true);
            static Timer timer = TIMER(TIMER_MILLISECONDS, CFG_LED_BILLOW_SPEED);
            if( !TIMER_RestartIfExpired(&timer) ) return;
            SInt08 brightness = BOARD_LedBrightness();
            if( rising )
            {   brightness++;
                if( brightness == 100 ) rising = false;
            }
            else
            {   brightness--;
                if( brightness == 0 ) rising = true;
            }
            BOARD_SetLedBrightness(brightness);
            break;
        case RS_DISABLING:
            // Radio wird ausgeschaltet
            SOUND_Start(SOUND_RADIO_OFF);
            DISPLAY_StartAnim(ANIM_RADIO_OFF);
            radio = RS_OFF;
            break;
        default:
            // Radio ist ausgeschaltet
            BOARD_EnableRelais(false);
            BOARD_SetLedBrightness(0);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  ProceedPPins(): Verfahrensfunktion zur Schaltung der frei Nutzbaren Pins P1 und P2. //
//  --------------																		//
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -																			//
//////////////////////////////////////////////////////////////////////////////////////////
static void
ProceedPPins(void)
{
	// Pin P1 wird gesetzt, wenn der aktuelle Tag ein freier Tag ist
	DateTime  dateTime;
    GetLocalDateTime(&dateTime);
	bool on = IsOffDay(&dateTime.date);    
    BOARD_EnableP1(on);
	// Pin 2 wird gesetzt, wenn Weckalarm ist
	BOARD_EnableP2(ALARM_IsAlarm());
}
		
//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  APPLICATION_Mode(): Liefert die aktuelle Betriebsart des Weckers.                   //
//  ------------------                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Betriebsmodus als Aufzählungstyp 'OperationMode'.                          //
//////////////////////////////////////////////////////////////////////////////////////////
OperationMode
APPLICATION_Mode(void)
{
    return mode;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  APPLICATION_ChangeModeTo(): Veranlasst  den Wechsel des Betriebsmodus. Der eigent-  //
//  --------------------------  liche Wechsel wird von der Verfahrensfunktion Proceed-  //
//  Mode() durchgeführt.                                                                //
//                                                                                      //
//  Eingabe: · newMode: Der neue Betriebsmodus als Aufzählungstyp 'OperationMode'.      //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_ChangeModeTo(OperationMode newMode)
{
    mode = newMode;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
APPLICATION_Key(void)
{
    return keyPressed;
}

//////////////////////////////////////////////////////////////////////////////////////////
SInt08
APPLICATION_Encoder(void)
{
    return encoderPosition;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
APPLICATION_UserInteraction(void)
{
    return APPLICATION_Key() || APPLICATION_Encoder() != 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
APPLICATION_RadioManuallyOn(void)
{
    return radio == RS_ENABLING || radio == RS_ON;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_WakeDisplay(void)
{
    wakeup = true;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
APPLICATION_IsHoliday(DTDayType dayType)
{
    GermanRegion region;
    STORING_Load(STO_GermanRegion, &region);
    return IsHoliday(region, dayType);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  System  //////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    DISPLAY_PrintLit(" OctEt7 ");
    SOUND_Start(SOUND_STARTUP);
    for( UInt08 brightness=0; brightness<30; brightness++ )
    {   static Timer timer = TIMER(TIMER_MILLISECONDS, 100);
        if( TIMER_RestartIfExpired(&timer) )
        {   DISPLAY_SetBrightness(brightness);
        }
        while( !TIMER_HasExpired(&timer) )
        {   BOARD_Proceed();
            DISPLAY_Proceed();
        }
    }
    APPLICATION_ChangeModeTo(OM_RECEIVING);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{	
    if( DISPLAY_CurrentAnim() == ANIM_NONE )
    {	static OperationMode  currentMode = (OperationMode)(-1);
		if( currentMode != mode )
		{   // Aufruf der Betriebsart-Einstiegsfunktion
			switch( mode )
			{   case OM_RECEIVING: EnterReceiveMode(); break;
				case OM_CLOCK:     EnterClockMode();   break;           
				case OM_MENU:      MENU_Enter();       break;
				case OM_SETTINGS:  SETTINGS_Enter();   break;
				default:           ;
			}
			currentMode = mode;
		}
		else
		{   // Aufruf der zu 'mode' passenden Verfahrensfunktion
			switch( mode )
			{   case OM_RECEIVING: ProceedReceiveMode(); break;
				case OM_CLOCK:     ProceedClockMode();   break;            
				case OM_MENU:      MENU_Proceed();       break;
				case OM_SETTINGS:  SETTINGS_Proceed();   break;
				default:           ;
			}
		}
	}
    ProceedUserInteraction();
    ProceedDisplaySaving();
    ProceedDisplayBrightness();
    ProceedDotSignaling();
    ProceedRadioSSR();
	ProceedPPins();
    ALARM_Proceed();
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
